"""
The `nntm.model_selection._split` module includes classes and
functions to split the data based on a preset strategy.
"""

# Author: Alexandre Gramfort <alexandre.gramfort@inria.fr>
#         Gael Varoquaux <gael.varoquaux@normalesup.org>
#         Olivier Grisel <olivier.grisel@ensta.org>
#         Raghav RV <rvraghav93@gmail.com>
#         Leandro Hermida <hermidal@cs.umd.edu>
#         Rodion Martynov <marrodon@gmail.com>
# License: BSD 3 clause
# Author: Timo Sutterer <hi@timo-sutterer.de>
# License: MIT

import logging
import numbers
import numpy as np
import pandas as pd
from sklearn.model_selection import BaseCrossValidator
from sklearn.utils import indexable
from ..utils.validation import _num_samples

logger = logging.getLogger(__name__)

__all__ = ["PurgedKFold"]


class PurgedKFold(BaseCrossValidator):
    """Purged K-Folds cross-validator

    Provides train/test indices to split data in train/test sets. Split
    dataset into k consecutive folds. Training observations overlapping
    in time with test observations are purged.

    Optionally, the eras that immediately follow the test set can be
    eliminated using the `embargo` argument.

    Data is assumed to be contiguous (shuffle=False).

    References
    ----------
    .. [1] `Marcos Lopez de Prado (2018). Advances in Financial Machine
            Learning. Chapter 7 (Cross-Validation in Finance).`_
    .. [2] `Super Massive Data Release: Deep Dive
            <https://forum.numer.ai/t/super-massive-data-release-deep-dive/4053>`_
    """

    def __init__(self, n_splits=5, target_days=20, embargo=None):
        if not isinstance(n_splits, numbers.Integral):
            raise ValueError(
                "The number of folds must be of Integral type. "
                f"`n_splits={n_splits}` of type {type(n_splits)} was passed."
            )
        n_splits = int(n_splits)

        if n_splits <= 1:
            raise ValueError(
                "k-fold cross-validation requires at least one "
                "train/test split by setting `n_splits=2` or more, "
                f"got `n_splits={n_splits}`."
            )

        if not isinstance(target_days, numbers.Integral):
            raise ValueError(
                "The number of target days must be of Integral type. "
                f"`target_days={target_days}` of type {type(target_days)} was passed."
            )
        target_days = int(target_days)

        if target_days % 5 != 0:
            raise ValueError(
                "The number of target days has to be a multiple of 5. "
                f"`target_days={target_days}` was passed."
            )

        if embargo:
            if not isinstance(embargo, float):
                raise ValueError(
                    "Embargo must be of float type. "
                    f"`embargo={embargo}` of type {type(embargo)} was passed."
                )

            if not 0.0 < embargo < 1.0:
                raise ValueError(
                    "Embargo must be between 0.0 and 1.0. "
                    f"`embargo={embargo}` was passed."
                )

        self.n_splits = n_splits
        self.target_days = target_days
        self.embargo = embargo

    def split(self, X, y=None, groups=None):
        """Generate indices to split data into training and test set.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.
        y : array-like of shape (n_samples,), default=None
            The target variable for supervised learning problems.
        groups : array-like of shape (n_samples,), default=None
            Eras for the samples used while splitting the dataset into
            train/test set. This parameter is not required when X is
            a pandas DataFrame containing an `era` column.
        Yields
        ------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """
        if isinstance(X, np.ndarray) and groups is None:
            raise ValueError("`groups` parameter is required when X is a numpy array")

        if isinstance(X, pd.DataFrame) and groups is None and "era" not in X.columns:
            raise ValueError("`groups` parameter is required when X has no era column")

        X, y, groups = indexable(X, y, groups)
        n_samples = _num_samples(X)
        if self.n_splits > n_samples:
            raise ValueError(
                (
                    f"Cannot have number of splits n_splits={self.n_splits} greater "
                    f"than the number of samples: n_samples={n_samples}."
                )
            )

        eras = np.fromiter(self._get_eras(X, groups=groups), dtype=int)
        target_weeks = self.target_days // 5
        eras_target_release = np.array([era + target_weeks - 1 for era in eras])

        embargo_era_count = 0
        if self.embargo:
            era_count = len(set(eras))
            embargo_era_count = int(round(era_count * self.embargo))

        indices = np.arange(_num_samples(X))
        for test_index_mask in self._iter_test_masks(X, y, groups):
            test_index = indices[test_index_mask]
            test_era_min = min(eras[test_index])
            test_era_max = max(eras[test_index])
            test_era_target_release_max = max(eras_target_release[test_index])

            train_index = indices[np.logical_not(test_index_mask)]
            for idx, train_era, train_era_target_release in zip(
                train_index, eras[train_index], eras_target_release[train_index]
            ):
                purge = not (
                    train_era_target_release < test_era_min
                    or train_era > test_era_target_release_max
                )
                embargo = test_era_max <= train_era <= test_era_max + embargo_era_count

                if purge or embargo:
                    train_index = train_index[train_index != idx]

            yield train_index, test_index

    def _iter_test_indices(self, X=None, y=None, groups=None):
        """Generates integer indices corresponding to test sets."""
        n_samples = _num_samples(X)
        indices = np.arange(n_samples)

        n_splits = self.n_splits

        # Fold sizes depend on n_samples (not n_eras)
        fold_sizes = np.full(n_splits, n_samples // n_splits, dtype=int)
        fold_sizes[: n_samples % n_splits] += 1
        current = 0
        for fold_size in fold_sizes:
            start, stop = current, current + fold_size
            yield indices[start:stop]
            current = stop

    def _get_eras(self, X, groups=None):
        """Generates integer eras."""
        eras = groups
        if groups is None:
            eras = X["era"].tolist()

        for era in eras:
            yield int(era)

    def get_n_splits(self, X=None, y=None, groups=None):
        """Returns the number of splitting iterations in the cross-validator
        Parameters
        ----------
        X : object
            Always ignored, exists for sklearn compatibility.
        y : object
            Always ignored, exists for sklearn compatibility.
        groups : object
            Always ignored, exists for sklearn compatibility.
        Returns
        -------
        n_splits : int
            Returns the number of splitting iterations in the cross-validator.
        """
        return self.n_splits



"""Numerai main tournament dataset.

The Numerai main tournament dataset is made of high quality financial
data that has been cleaned, regularized and obfuscated.

The original dataset is available from

    https://numerai-public-datasets.s3-us-west-2.amazonaws.com/

A detailed description of the dataset is available from

    https://forum.numer.ai/t/super-massive-data-release-deep-dive/4053


The dataset is divided into:

Training data:
--------------
- 2412105 samples over 574 eras (1-574)
- 1050 features
- 20 different targets

Test data:
----------
- 1407586 samples over 278 eras (575-852)
- 1050 features
- no target values provided

Validation data:
----------------
- 539658 samples over 105 eras (857-961)
- 1050 features
- 20 different targets

Live data:
----------
- changes weekly
- number of samples may vary, one era
- 1050 features
- no target values provided

Tournament data:
----------------
- test data and live data

More data may be added from time to time.

References
----------

https://docs.numer.ai/tournament/learn
https://forum.numer.ai/t/super-massive-data-release-deep-dive/4053
https://github.com/numerai/example-scripts/blob/0a8c4f764a3aee3b7c1709058dd1488b26bd5f01/analysis_and_tips.ipynb
https://github.com/uuazed/numerapi

"""

import json
import logging
from os.path import exists, splitext
from os import makedirs, remove
from typing import List

import pandas as pd
from numerapi import NumerAPI
from sklearn.datasets import get_data_home
from sklearn.utils import Bunch

logger = logging.getLogger(__name__)


def _get_numerai_fetcher(filename_float32, filename_int8, name, has_rounds=False):
    """Return fetch function for passed files"""

    def fetch_numerai(
        *,
        data_home=None,
        download_if_missing=True,
        return_X_y=False,
        target="target",
        as_frame=False,
        columns=None,
        int8=True,
        round_num=None,
    ):
        # Make sure required columns will be read
        if columns:
            columns = list(set(columns + ["era", "data_type"]))

        # Get round (if dataset supports rounds)
        napi = NumerAPI()
        if round_num and not has_rounds:
            raise ValueError("`round_num` given for a dataset without rounds.")
        if not round_num and has_rounds:
            round_num = napi.get_current_round()
            logger.info(f"Using current round={round_num}")

        # Get file locations
        data_home = get_data_home(data_home=data_home)
        if not exists(data_home):
            makedirs(data_home)
        filename = filename_float32
        if int8:
            filename = filename_int8
        filepath = "/".join([data_home, filename])
        if round_num:
            suffix = "_" + str(round_num)
            filename_with_suffix = _add_filename_suffix(filename, suffix)
            filepath = "/".join([data_home, filename_with_suffix])
        else:
            filepath = "/".join([data_home, filename])

        # Download and read dataset
        if not exists(filepath):
            if not download_if_missing:
                raise IOError("Data not found and `download_if_missing` is False")

            logger.info(f"Downloading Numerai {name} data to {filepath}")
            napi.download_dataset(filename, dest_path=filepath)

            df = pd.read_parquet(filepath, columns=columns)
            remove(filepath)
        else:
            df = pd.read_parquet(filepath, columns=columns)

        # Get feature and target columns
        feature_names = _get_feature_names(df)
        target_names = _get_target_names(df)

        # Enforce expected data type for features
        dtype = _get_dtype(is_int8=int8)
        df[feature_names] = df[feature_names].astype(dtype)

        X = df[feature_names]
        # One entry per target
        target_dict = {tn: df[tn] for tn in target_names}
        # All targets in one DataFrame
        targets = df[target_names]

        id_ = df.index
        era = df.era
        data_type = df.data_type

        if not as_frame:
            # Convert to numpy
            X = X.to_numpy(dtype=dtype)
            target_dict = {k: v.to_numpy() for k, v in target_dict.items()}
            targets = targets.to_numpy()
            id_ = id_.to_numpy()
            era = era.to_numpy()
            data_type = data_type.to_numpy()
            df = None

        if return_X_y:
            return X, target_dict[target]

        return Bunch(
            data=X,
            **target_dict,
            targets=targets,
            id=id_,
            era=era,
            data_type=data_type,
            feature_names=feature_names,
            target_names=target_names,
            int8=int8,
            DESCR=f"Numerai main tournament: {name} data",
            frame=df,
        )

    return fetch_numerai


def fetch_numerai_training(*args, **kwargs):
    """Load the Numerai training dataset.

    Parameters
    ----------
    data_home : str, default=None
        Specify another download and cache folder for the datasets. By default
        all data is stored in `~/scikit_learn_data` subfolders.

    download_if_missing : bool, default=True
        If False, raise an IOError if the data is not locally available
        instead of trying to download the data from the source bucket.

    return_X_y : bool, default=False
        If True, returns `(data.data, data.target)` instead of a Bunch
        object. A custom target column can be selected through the
        `target` parameter.

    target : str, default="target"
        Target column to return as `y` when `return_X_y=True`.

    as_frame : bool, default=False
        If True, `data` and `targets` are pandas DataFrames. `target`,
        `target_<name>`, `id`, `era` and `data_type` are pandas Series.
        `frame` will be given.

    columns : list, default=None
        If not None, only these columns will be read from the file.
        `index`, `era` and `data_type` columns are always read.

    int8 : bool, default=True
        If True, the feature columns will use the `int8` data type
        instead of `float32`. Target columns are always `float32`.

    Returns
    -------
    dataset : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.

        data : {ndarray, DataFrame} of shape (2412105, 1050)
            Each row corresponding to the `feature_names` in order.
            When `as_frame=True`, `data` is a pandas DataFrame.

        target : {ndarray, Series} of shape (2412105,)
            When `as_frame=True`, `target` is a pandas Series.

        targets : {ndarray, DataFrame} of shape (2412105, 21)
            When `as_frame=True`, `targets` is a pandas DataFrame.

        target_<name> : {ndarray, Series} of shape (2412105,)
            See `target_names` for available targets.
            When `as_frame=True`, `target_<name>` is a pandas Series.

        id : {ndarray, Series} of shape (2412105,)
            `id` of each row in `data`.
            When `as_frame=True`, `id` is a pandas Series.

        era : {ndarray, Series} of shape (2412105,)
            `era` of each row in `data`.
            When `as_frame=True`, `era` is a pandas Series.

        data_type : {ndarray, Series} of shape (2412105,)
            `data_type` of each row in `data`.
            When `as_frame=True`, `data_type` is a pandas Series.

        feature_names : list of length 1050
            List of ordered feature names used in the dataset.

        target_names : list of length 21
            List of ordered target names used in the dataset.

        int8 : bool
            True when features use `int8` data type.

        DESCR : string
            Description of the dataset.

        frame : DataFrame if `as_frame=True`
            Only present when `as_frame=True`. Pandas DataFrame with
            `era`, `data_type`, features and targets.

        (data, target) : tuple if `return_X_y=True`
            Only present when `return_X_y=True`
            `target` corresponds to the column set by the `target` attribute.
    """
    fetcher = _get_numerai_fetcher(
        "numerai_training_data.parquet",
        "numerai_training_data_int8.parquet",
        "training",
    )
    return fetcher(*args, **kwargs)


def fetch_numerai_test(*args, **kwargs):
    """Load the Numerai test dataset.

    Parameters
    ----------
    data_home : str, default=None
        Specify another download and cache folder for the datasets. By default
        all data is stored in `~/scikit_learn_data` subfolders.

    download_if_missing : bool, default=True
        If False, raise an IOError if the data is not locally available
        instead of trying to download the data from the source bucket.

    return_X_y : bool, default=False
        If True, returns `(data.data, data.target)` instead of a Bunch
        object. A custom target column can be selected through the
        `target` parameter.

    target : str, default="target"
        Target column to return as `y` when `return_X_y=True`.

    as_frame : bool, default=False
        If True, `data` and `targets` are pandas DataFrames. `target`,
        `target_<name>`, `id`, `era` and `data_type` are pandas Series.
        `frame` will be given.

    columns : list, default=None
        If not None, only these columns will be read from the file.
        `index`, `era` and `data_type` columns are always read.

    int8 : bool, default=True
        If True, the feature columns will use the `int8` data type
        instead of `float32`. Target columns are always `float32`.

    Returns
    -------
    dataset : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.

        data : {ndarray, DataFrame} of shape (1407586, 1050)
            Each row corresponding to the `feature_names` in order.
            When `as_frame=True`, `data` is a pandas DataFrame.

        target : {ndarray, Series} of shape (1407586,)
            When `as_frame=True`, `target` is a pandas Series.

        targets : {ndarray, DataFrame} of shape (1407586, 21)
            When `as_frame=True`, `targets` is a pandas DataFrame.

        target_<name> : {ndarray, Series} of shape (1407586,)
            See `target_names` for available targets.
            When `as_frame=True`, `target_<name>` is a pandas Series.

        id : {ndarray, Series} of shape (1407586,)
            `id` of each row in `data`.
            When `as_frame=True`, `id` is a pandas Series.

        era : {ndarray, Series} of shape (1407586,)
            `era` of each row in `data`.
            When `as_frame=True`, `era` is a pandas Series.

        data_type : {ndarray, Series} of shape (1407586,)
            `data_type` of each row in `data`.
            When `as_frame=True`, `data_type` is a pandas Series.

        feature_names : list of length 1050
            List of ordered feature names used in the dataset.

        target_names : list of length 21
            List of ordered target names used in the dataset.

        int8 : bool
            True when features use `int8` data type.

        DESCR : string
            Description of the dataset.

        frame : DataFrame if `as_frame=True`
            Only present when `as_frame=True`. Pandas DataFrame with
            `era`, `data_type`, features and targets.

        (data, target) : tuple if `return_X_y=True`
            Only present when `return_X_y=True`
            `target` corresponds to the column set by the `target` attribute.
    """
    fetcher = _get_numerai_fetcher(
        "numerai_test_data.parquet",
        "numerai_test_data_int8.parquet",
        "test",
    )
    return fetcher(*args, **kwargs)


def fetch_numerai_validation(*args, **kwargs):
    """Load the Numerai validation dataset.

    Parameters
    ----------
    data_home : str, default=None
        Specify another download and cache folder for the datasets. By
        default all data is stored in `~/scikit_learn_data` subfolders.

    download_if_missing : bool, default=True
        If False, raise an IOError if the data is not locally available
        instead of trying to download the data from the source bucket.

    return_X_y : bool, default=False
        If True, returns `(data.data, data.target)` instead of a Bunch
        object. A custom target column can be selected through the
        `target` parameter.

    target : str, default="target"
        Target column to return as `y` when `return_X_y=True`.

    as_frame : bool, default=False
        If True, `data` and `targets` are pandas DataFrames. `target`,
        `target_<name>`, `id`, `era` and `data_type` are pandas Series.
        `frame` will be given.

    columns : list, default=None
        If not None, only these columns will be read from the file.
        `index`, `era` and `data_type` columns are always read.

    int8 : bool, default=True
        If True, the feature columns will use the `int8` data type
        instead of `float32`. Target columns are always `float32`.

    Returns
    -------
    dataset : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.

        data : {ndarray, DataFrame} of shape (539658, 1050)
            Each row corresponding to the `feature_names` in order.
            When `as_frame=True`, `data` is a pandas DataFrame.

        target : {ndarray, Series} of shape (539658,)
            When `as_frame=True`, `target` is a pandas Series.

        targets : {ndarray, DataFrame} of shape (539658, 21)
            When `as_frame=True`, `targets` is a pandas DataFrame.

        target_<name> : {ndarray, Series} of shape (539658,)
            See `target_names` for available targets.
            When `as_frame=True`, `target_<name>` is a pandas Series.

        id : {ndarray, Series} of shape (539658,)
            `id` of each row in `data`.
            When `as_frame=True`, `id` is a pandas Series.

        era : {ndarray, Series} of shape (539658,)
            `era` of each row in `data`.
            When `as_frame=True`, `era` is a pandas Series.

        data_type : {ndarray, Series} of shape (539658,)
            `data_type` of each row in `data`.
            When `as_frame=True`, `data_type` is a pandas Series.

        feature_names : list of length 1050
            List of ordered feature names used in the dataset.

        target_names : list of length 21
            List of ordered target names used in the dataset.

        int8 : bool
            True when features use `int8` data type.

        DESCR : string
            Description of the dataset.

        frame : DataFrame if `as_frame=True`
            Only present when `as_frame=True`. Pandas DataFrame with
            `era`, `data_type`, features and targets.

        (data, target) : tuple if `return_X_y=True`
            Only present when `return_X_y=True`
            `target` corresponds to the column set by the `target`
            attribute.
    """
    fetcher = _get_numerai_fetcher(
        "numerai_validation_data.parquet",
        "numerai_validation_data_int8.parquet",
        "validation",
    )
    return fetcher(*args, **kwargs)


def fetch_numerai_live(*args, **kwargs):
    """Load the Numerai live dataset.

    Parameters
    ----------
    data_home : str, default=None
        Specify another download and cache folder for the datasets. By
        default all data is stored in `~/scikit_learn_data` subfolders.

    download_if_missing : bool, default=True
        If False, raise an IOError if the data is not locally available
        instead of trying to download the data from the source bucket.

    return_X_y : bool, default=False
        If True, returns `(data.data, data.target)` instead of a Bunch
        object. A custom target column can be selected through the
        `target` parameter.

    target : str, default="target"
        Target column to return as `y` when `return_X_y=True`.

    as_frame : bool, default=False
        If True, `data` and `targets` are pandas DataFrames. `target`,
        `target_<name>`, `id`, `era` and `data_type` are pandas Series.
        `frame` will be given.

    columns : list, default=None
        If not None, only these columns will be read from the file.
        `index`, `era` and `data_type` columns are always read.

    int8 : bool, default=True
        If True, the feature columns will use the `int8` data type
        instead of `float32`. Target columns are always `float32`.

    round_num : int, default=None
        Tournament round to download. If None, current round will be
        downloaded.

    Returns
    -------
    dataset : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.

        data : {ndarray, DataFrame} of shape (n_samples, 1050)
            Each row corresponding to the `feature_names` in order.
            When `as_frame=True`, `data` is a pandas DataFrame.

        target : {ndarray, Series} of shape (n_samples,)
            When `as_frame=True`, `target` is a pandas Series.

        targets : {ndarray, DataFrame} of shape (n_samples, 21)
            When `as_frame=True`, `targets` is a pandas DataFrame.

        target_<name> : {ndarray, Series} of shape (n_samples,)
            See `target_names` for available targets.
            When `as_frame=True`, `target_<name>` is a pandas Series.

        id : {ndarray, Series} of shape (n_samples,)
            `id` of each row in `data`.
            When `as_frame=True`, `id` is a pandas Series.

        era : {ndarray, Series} of shape (n_samples,)
            `era` of each row in `data`.
            When `as_frame=True`, `era` is a pandas Series.

        data_type : {ndarray, Series} of shape (n_samples,)
            `data_type` of each row in `data`.
            When `as_frame=True`, `data_type` is a pandas Series.

        feature_names : list of length 1050
            List of ordered feature names used in the dataset.

        target_names : list of length 21
            List of ordered target names used in the dataset.

        int8 : bool
            True when features use `int8` data type.

        DESCR : string
            Description of the dataset.

        frame : DataFrame if `as_frame=True`
            Only present when `as_frame=True`. Pandas DataFrame with
            `era`, `data_type`, features and targets.

        (data, target) : tuple if `return_X_y=True`
            Only present when `return_X_y=True`
            `target` corresponds to the column set by the `target`
            attribute.

    Notes
    -----
    Data changes weekly.
    """
    fetcher = _get_numerai_fetcher(
        "numerai_live_data.parquet",
        "numerai_live_data_int8.parquet",
        "live",
        has_rounds=True,
    )
    return fetcher(*args, **kwargs)


def fetch_numerai_tournament(*args, **kwargs):
    """Load the Numerai tournament dataset.

    Parameters
    ----------
    data_home : str, default=None
        Specify another download and cache folder for the datasets. By
        default all data is stored in `~/scikit_learn_data` subfolders.

    download_if_missing : bool, default=True
        If False, raise an IOError if the data is not locally available
        instead of trying to download the data from the source bucket.

    return_X_y : bool, default=False
        If True, returns `(data.data, data.target)` instead of a Bunch
        object. A custom target column can be selected through the
        `target` parameter.

    target : str, default="target"
        Target column to return as `y` when `return_X_y=True`.

    as_frame : bool, default=False
        If True, `data` and `targets` are pandas DataFrames. `target`,
        `target_<name>`, `id`, `era` and `data_type` are pandas Series.
        `frame` will be given.

    columns : list, default=None
        If not None, only these columns will be read from the file.
        `index`, `era` and `data_type` columns are always read.

    int8 : bool, default=True
        If True, the feature columns will use the `int8` data type
        instead of `float32`. Target columns are always `float32`.

    round_num : int, default=None
        Tournament round to download. If None, current round will be
        downloaded.

    Returns
    -------
    dataset : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.

        data : {ndarray, DataFrame} of shape (n_samples, 1050)
            Each row corresponding to the `feature_names` in order.
            When `as_frame=True`, `data` is a pandas DataFrame.

        target : {ndarray, Series} of shape (n_samples,)
            When `as_frame=True`, `target` is a pandas Series.

        targets : {ndarray, DataFrame} of shape (n_samples, 21)
            When `as_frame=True`, `targets` is a pandas DataFrame.

        target_<name> : {ndarray, Series} of shape (n_samples,)
            See `target_names` for available targets.
            When `as_frame=True`, `target_<name>` is a pandas Series.

        id : {ndarray, Series} of shape (n_samples,)
            `id` of each row in `data`.
            When `as_frame=True`, `id` is a pandas Series.

        era : {ndarray, Series} of shape (n_samples,)
            `era` of each row in `data`.
            When `as_frame=True`, `era` is a pandas Series.

        data_type : {ndarray, Series} of shape (n_samples,)
            `data_type` of each row in `data`.
            When `as_frame=True`, `data_type` is a pandas Series.

        feature_names : list of length 1050
            List of ordered feature names used in the dataset.

        target_names : list of length 21
            List of ordered target names used in the dataset.

        int8 : bool
            True when features use `int8` data type.

        DESCR : string
            Description of the dataset.

        frame : DataFrame if `as_frame=True`
            Only present when `as_frame=True`. Pandas DataFrame with
            `era`, `data_type`, features and targets.

        (data, target) : tuple if `return_X_y=True`
            Only present when `return_X_y=True`
            `target` corresponds to the column set by the `target`
            attribute.

    Notes
    -----
    Data changes weekly.
    """
    fetcher = _get_numerai_fetcher(
        "numerai_tournament_data.parquet",
        "numerai_tournament_data_int8.parquet",
        "tournament",
        has_rounds=True,
    )
    return fetcher(*args, **kwargs)


def _get_numerai_prediction_fetcher(fname, name, has_rounds=False):
    """Return prediction fetch function for passed files"""

    def fetch_numerai_predictions(
        *,
        data_home=None,
        download_if_missing=True,
        return_y=False,
        as_frame=False,
        round_num=None,
    ):
        # Get round (if predictions support rounds)
        napi = NumerAPI()
        if round_num and not has_rounds:
            raise ValueError("`round_num` given for predictions without rounds.")
        if not round_num and has_rounds:
            round_num = napi.get_current_round()
            logger.info(f"Using current round={round_num}")

        # Get file locations
        data_home = get_data_home(data_home=data_home)
        filename = fname
        if not exists(data_home):
            makedirs(data_home)
        if round_num:
            suffix = "_" + str(round_num)
            filename_with_suffix = _add_filename_suffix(filename, suffix)
            filepath = "/".join([data_home, filename_with_suffix])
        else:
            filepath = "/".join([data_home, filename])

        # Download and read dataset
        if not exists(filepath):
            if not download_if_missing:
                raise IOError("Data not found and `download_if_missing` is False")

            logger.info(f"Downloading Numerai {name} predictions to {filepath}")
            napi.download_dataset(filename, dest_path=filepath)

            df = pd.read_parquet(filepath)
            remove(filepath)
        else:
            df = pd.read_parquet(filepath)

        y = df.prediction
        id_ = df.index

        if not as_frame:
            # Convert to numpy
            y = y.to_numpy()
            id_ = id_.to_numpy()
            df = None

        if return_y:
            return y

        return Bunch(
            prediction=y,
            id=id_,
            DESCR=f"Numerai main tournament: {name} predictions",
            frame=df,
        )

    return fetch_numerai_predictions


def fetch_numerai_example_predictions(*args, **kwargs):
    """Load the Numerai example predictions.

    Parameters
    ----------
    data_home : str, default=None
        Specify another download and cache folder for the datasets. By
        default all data is stored in `~/scikit_learn_data` subfolders.

    download_if_missing : bool, default=True
        If False, raise an IOError if the data is not locally available
        instead of trying to download the data from the source bucket.

    return_y : bool, default=False.
        If True, returns `prediction` instead of a Bunch object.

    as_frame : bool, default=False
        If True, `prediction` and `id` are pandas Series. `frame` will
        be given.

    round_num : int, default=None
        Prediction round to download. If None, current round will be
        downloaded.

    Returns
    -------
    dataset : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.

        prediction : {ndarray, Series} of shape (n_samples,)
            When `as_frame=True`, `prediction` is a pandas Series.

        id : {ndarray, Series} of shape (n_samples,)
            `id` of each `prediction` row.
            When `as_frame=True`, `id` is a pandas Series.

        frame : DataFrame if `as_frame=True`
            Only present when `as_frame=True`. Pandas DataFrame with
            `prediction`.

        y : {ndarray, Series} of shape (n_samples,) if `return_y=True`
            Only present when `return_y=True`. When `as_frame=True`,
            `y` is a pandas Series.

    Notes
    -----
    Data changes weekly.
    """
    fetcher = _get_numerai_prediction_fetcher(
        "example_predictions.parquet",
        "example",
        has_rounds=True,
    )
    return fetcher(*args, **kwargs)


def fetch_numerai_example_validation_predictions(*args, **kwargs):
    """Load the Numerai example validation predictions.

    Parameters
    ----------
    data_home : str, default=None
        Specify another download and cache folder for the datasets. By
        default all data is stored in `~/scikit_learn_data` subfolders.

    download_if_missing : bool, default=True
        If False, raise an IOError if the data is not locally available
        instead of trying to download the data from the source bucket.

    return_y : bool, default=False.
        If True, returns `prediction` instead of a Bunch object.

    as_frame : bool, default=False
        If True, `prediction` and `id` are pandas Series. `frame` will
        be given.

    Returns
    -------
    dataset : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.

        prediction : {ndarray, Series} of shape (539658,)
            When `as_frame=True`, `prediction` is a pandas Series.

        id : {ndarray, Series} of shape (539658,)
            `id` of each `prediction` row.
            When `as_frame=True`, `id` is a pandas Series.

        frame : DataFrame if `as_frame=True`
            Only present when `as_frame=True`. Pandas DataFrame with
            `prediction`.

        y : {ndarray, Series} of shape (539658,) if `return_y=True`
            Only present when `return_y=True`. When `as_frame=True`,
            `y` is a pandas Series.
    """
    fetcher = _get_numerai_prediction_fetcher(
        "example_validation_predictions.parquet",
        "example validation",
    )
    return fetcher(*args, **kwargs)


def fetch_numerai_feature_metadata(*, data_home=None, download_if_missing=True):
    """Load the Numerai feature metadata.

    Parameters
    ----------
    data_home : str, default=None
        Specify another download and cache folder for the feature
        metadata. By default all data is stored in
        `~/scikit_learn_data` subfolders.

    download_if_missing : bool, default=True
        If False, raise an IOError if the data is not locally available
        instead of trying to download the data from the source bucket.

    Returns
    -------
    feature_metadata : dict
        Dictionary with the following keys.

        feature_sets : dict
            Dictionary containing lists of feature names as values.

        feature_stats : dict
            Dictionary with feature names as keys and a dictionary
            of different statistics about each feature as values.

    References
    ----------

    https://forum.numer.ai/t/october-2021-updates/4384

    """
    # Get file locations
    data_home = get_data_home(data_home=data_home)
    filename = "features.json"
    if not exists(data_home):
        makedirs(data_home)
    filepath = "/".join([data_home, filename])

    # Download and read dataset
    if not exists(filepath):
        if not download_if_missing:
            raise IOError("Data not found and `download_if_missing` is False")

        logger.info(f"Downloading Numerai feature metadata to {filepath}")
        napi = NumerAPI()
        napi.download_dataset(filename, dest_path=filepath)

        feature_metadata_dict = _read_json_file(filepath)
        remove(filepath)
    else:
        feature_metadata_dict = _read_json_file(filepath)

    return feature_metadata_dict


def _get_feature_names(df: pd.DataFrame) -> List[str]:
    """Get list of `df`s feature column names."""
    return [c for c in df if c.startswith("feature_")]


def _get_target_names(df: pd.DataFrame) -> List[str]:
    """Get list of `df`s target column names."""
    return [c for c in df if c.startswith("target")]


def _get_dtype(is_int8: bool) -> str:
    """Return data type. Either int8 or float32."""
    if is_int8:
        return "int8"
    else:
        return "float32"


def _add_filename_suffix(filepath: str, suffix: str) -> str:
    """Insert suffix between current filename and extension."""
    filepath, file_extension = splitext(filepath)
    return filepath + suffix + file_extension


def _read_json_file(filepath: str) -> dict:
    """Return content of json file at filepath."""
    with open(filepath) as json_file:
        content = json.load(json_file)

    return content



"""Test the split module"""
import secrets
import pytest
import numpy as np
import pandas as pd
from nntm.model_selection import PurgedKFold

ERAS_TOTAL = 100
SAMPLES_PER_ERA = 3
SAMPLES_TOTAL = ERAS_TOTAL * SAMPLES_PER_ERA
eras = [str(i + 1) for i in range(ERAS_TOTAL) for _ in range(SAMPLES_PER_ERA)]
feat_a = [1 for _ in range(SAMPLES_TOTAL)]
feat_b = [3 for _ in range(SAMPLES_TOTAL)]
unique_ids = [secrets.token_hex(nbytes=8) for _ in range(SAMPLES_TOTAL)]
X_np = np.array([feat_a, feat_b]).T
y_np = np.array([0.5 for _ in range(SAMPLES_TOTAL)])
X_df = pd.DataFrame(
    {
        "id": unique_ids,
        "era": eras,
        "feature_a": feat_a,
        "feature_b": feat_b,
    }
).set_index("id")
y_df = pd.Series([0.5 for _ in range(SAMPLES_TOTAL)])


def test_purgedkfold_params():
    # Check defaults
    pkfold = PurgedKFold()
    assert str(pkfold) == "PurgedKFold(embargo=None, n_splits=5, target_days=20)"
    assert pkfold.get_n_splits() == 5
    # Let's use custom values
    pkfold = PurgedKFold(n_splits=9, target_days=60)
    assert str(pkfold) == "PurgedKFold(embargo=None, n_splits=9, target_days=60)"
    assert pkfold.get_n_splits() == 9
    # n_splits has to be an integer > 1
    with pytest.raises(ValueError):
        PurgedKFold(n_splits=5.2)
    with pytest.raises(ValueError):
        PurgedKFold(n_splits=1)
    # target_days has to be and integer and multiple of 5
    with pytest.raises(ValueError):
        PurgedKFold(target_days=60.0)
    with pytest.raises(ValueError):
        PurgedKFold(target_days=19)
    # embargo has to be a float between 0.0 and 1.0
    with pytest.raises(ValueError):
        PurgedKFold(embargo=5)
    with pytest.raises(ValueError):
        PurgedKFold(embargo=1.3)


def test_purgedkfold_np():
    # When X is a numpy array, the split method requires a groups
    # parameter to know what has to be purged
    pkfold = PurgedKFold()
    with pytest.raises(ValueError):
        list(pkfold.split(X=X_np))

    # Test number of purged samples
    check_purged_splits(X_np, y_np, eras)


def test_purgedkfold_df():
    # When X is a pandas dataframe that has no "era" column, the split
    # method requires a group parameter to know what has to be purged
    pkfold = PurgedKFold()
    X_df_without_era_col = X_df.drop(columns="era")
    with pytest.raises(ValueError):
        list(pkfold.split(X=X_df_without_era_col))

    # Test number of purged samples
    check_purged_splits(X_df, y_df, None)


def test_purgedkfold_embargo():
    eras_int = [int(era) for era in eras]
    era_count = len(set(eras_int))
    era_max = max(eras_int)

    for target_days, purge_n_eras in [(20, 3), (60, 11)]:
        for embargo in [0.01, 0.05, 0.1, 0.2, 0.5]:
            for n_splits in [2, 5, 10]:
                pkfold_embargo = PurgedKFold(
                    n_splits=n_splits, target_days=target_days, embargo=embargo
                )
                pkfold = PurgedKFold(n_splits=n_splits, target_days=target_days)
                for (train_idx_embargo, test_idx_embargo), (train_idx, test_idx) in zip(
                    pkfold_embargo.split(X=X_df), pkfold.split(X=X_df)
                ):
                    # Test data is not affected
                    assert np.array_equal(test_idx_embargo, test_idx)

                    # Embargo applies only after test set
                    test_eras = get_int_era_set(X_df, test_idx)
                    if era_max in test_eras:
                        # No embargo if test set is the last split
                        assert np.array_equal(train_idx_embargo, train_idx)
                    else:
                        # Make sure the right amount of eras has been left out
                        train_eras = get_int_era_set(X_df, train_idx)
                        emb_train_eras = get_int_era_set(X_df, train_idx_embargo)
                        # Embargo is given as a percentage of all eras
                        embargo_only = int(round(embargo * era_count))
                        # Embargo overlaps with purge
                        embargo_only = max(0, embargo_only - purge_n_eras)
                        # Embargo can only remove as much as exists
                        trailing_eras = [e for e in train_eras if e > max(test_eras)]
                        embargo_only = min(len(trailing_eras), embargo_only)

                        assert len(train_eras) == len(emb_train_eras) + embargo_only


def check_purged_splits(X, y, groups):
    """Make sure the correct amount of samples gets purged."""
    n_samples = len(X)

    # https://forum.numer.ai/t/super-massive-data-release-deep-dive/4053
    # You need to sample every 4th era to get non-overlapping eras
    # with the 20 day targets, but every 12th era to get
    # non-overlapping eras with the 60 day targets
    for target_days, purge_n_eras in [(20, 3), (60, 11)]:
        for n_splits in [2, 5, 10, 99]:
            # Nothing should be purged from the test set
            expected_test_set_length = np.full(
                n_splits, n_samples // n_splits, dtype=int
            )
            expected_test_set_length[: n_samples % n_splits] += 1

            pkfold = PurgedKFold(n_splits=n_splits, target_days=target_days)
            for (train_idx, test_idx), test_set_length in zip(
                pkfold.split(X=X, y=y, groups=groups), expected_test_set_length
            ):
                assert test_set_length > 0
                assert len(test_idx) == test_set_length

                test_era_set = get_int_era_set(X, test_idx, groups)
                test_era_min, test_era_max = min(test_era_set), max(test_era_set)
                train_era_set = get_int_era_set(X, train_idx, groups)

                # There is no intersection between test and train eras
                assert test_era_set.intersection(train_era_set) == set()

                # Eras immediately before and after test set were purged
                for era in range(test_era_min - purge_n_eras, test_era_min + 1):
                    assert era not in train_era_set
                for era in range(test_era_max, test_era_max + purge_n_eras + 1):
                    assert era not in train_era_set

                # Not too many eras were purged
                eras_left_count = len(test_era_set.union(train_era_set))
                eras_left_expected = ERAS_TOTAL - (purge_n_eras * 2)
                assert eras_left_count >= eras_left_expected


def get_int_era_set(X, idx, groups=None):
    """Return set of eras from samples under idx.
    Eras are converted to int.
    """
    if groups:
        groups = np.array(groups)
        era_list = groups[idx]
    else:
        X_sub = X.iloc[idx]
        era_list = X_sub.era

    return set([int(era) for era in era_list])



"""Utilities for input validation"""

# Authors: Olivier Grisel
#          Gael Varoquaux
#          Andreas Mueller
#          Lars Buitinck
#          Alexandre Gramfort
#          Nicolas Tresegnie
#          Sylvain Marie
# License: BSD 3 clause

import numbers
import numpy as np


def _num_samples(x):
    """Return number of samples in array-like x."""
    message = "Expected sequence or array-like, got %s" % type(x)
    if hasattr(x, "fit") and callable(x.fit):
        # Don't get num_samples from an ensembles length!
        raise TypeError(message)

    if not hasattr(x, "__len__") and not hasattr(x, "shape"):
        if hasattr(x, "__array__"):
            x = np.asarray(x)
        else:
            raise TypeError(message)

    if hasattr(x, "shape") and x.shape is not None:
        if len(x.shape) == 0:
            raise TypeError(
                "Singleton array %r cannot be considered a valid collection." % x
            )
        # Check that shape is returning an integer or default to len
        # Dask dataframes may not return numeric shape[0] value
        if isinstance(x.shape[0], numbers.Integral):
            return x.shape[0]

    try:
        return len(x)
    except TypeError as type_error:
        raise TypeError(message) from type_error



"""Numerai main tournament metadata.

Convenience module defining some of the more static variables of the
Numerai main tournament.

The tournament is constantly evolving. Metadata from this module might
be outdated. Retrieving the most recent information using the `fetch_`
functions of the `nntm.datasets` module should be preferred.

References
----------

https://forum.numer.ai/t/super-massive-data-release-deep-dive/4053
https://forum.numer.ai/t/october-2021-updates/4384
"""

TARGET_NAMES_UNIQUE = [
    "target_nomi_20",
    "target_nomi_60",
    "target_jerome_20",
    "target_jerome_60",
    "target_janet_20",
    "target_janet_60",
    "target_ben_20",
    "target_ben_60",
    "target_alan_20",
    "target_alan_60",
    "target_paul_20",
    "target_paul_60",
    "target_george_20",
    "target_george_60",
    "target_william_20",
    "target_william_60",
    "target_arthur_20",
    "target_arthur_60",
    "target_thomas_20",
    "target_thomas_60",
]
TARGET_NAMES = ["target"] + TARGET_NAMES_UNIQUE

FEATURE_NAMES_LEGACY = [
    "feature_unco_terefah_thirster",
    "feature_enlightening_mirthful_laurencin",
    "feature_tarry_meet_chapel",
    "feature_himyarite_tetragonal_deceit",
    "feature_dovetailed_winy_hanaper",
    "feature_tortured_arsenical_arable",
    "feature_unamazed_tumular_photomicrograph",
    "feature_unsurveyed_boyish_aleph",
    "feature_phellogenetic_vibrational_jocelyn",
    "feature_seemlier_reorient_monandry",
    "feature_unmalleable_resistant_kingston",
    "feature_lordly_lamellicorn_buxtehude",
    "feature_antipodal_unable_thievery",
    "feature_hypersonic_volcanological_footwear",
    "feature_unspotted_practiced_gland",
    "feature_casuistic_barbarian_monochromy",
    "feature_torturesome_estimable_preferrer",
    "feature_unlawful_superintendent_brunet",
    "feature_multilinear_sharpened_mouse",
    "feature_padded_peripteral_pericranium",
    "feature_antisubmarine_foregoing_cryosurgery",
    "feature_spagyric_echt_alum",
    "feature_epidermic_scruffiest_prosperity",
    "feature_chaldean_vixenly_propylite",
    "feature_emmetropic_heraclitean_conducting",
    "feature_telephonic_shakable_bollock",
    "feature_unapplicable_jerkiest_klemperer",
    "feature_perceivable_gasiform_psammite",
    "feature_unrelieved_rawish_cement",
    "feature_casemated_ibsenian_grantee",
    "feature_fragrant_fifteen_brian",
    "feature_apomictical_motorized_vaporisation",
    "feature_hysteric_mechanized_recklinghausen",
    "feature_reduplicate_conoid_albite",
    "feature_softish_unseparated_caudex",
    "feature_faustian_unventilated_lackluster",
    "feature_greedier_favorable_enthymeme",
    "feature_rural_inquisitional_trotline",
    "feature_mined_game_curse",
    "feature_smoggy_niftiest_lunch",
    "feature_paramagnetic_complex_gish",
    "feature_chaotic_granitoid_theist",
    "feature_basaltic_arid_scallion",
    "feature_unrated_intact_balmoral",
    "feature_revealable_aeonian_elvira",
    "feature_desiderative_commiserative_epizoa",
    "feature_subdued_spiffier_kano",
    "feature_flavourful_seismic_erica",
    "feature_doggish_whacking_headscarf",
    "feature_palatalized_unsucceeded_induration",
    "feature_conceding_ingrate_tablespoonful",
    "feature_misanthropic_knurliest_freebooty",
    "feature_inhibited_snowiest_drawing",
    "feature_induplicate_hoarse_disbursement",
    "feature_uninclosed_handcrafted_springing",
    "feature_bleeding_arabesque_pneuma",
    "feature_incitant_trochoidal_oculist",
    "feature_cairned_fumiest_ordaining",
    "feature_huskiest_compartmental_jacquerie",
    "feature_endangered_unthreaded_firebrick",
    "feature_epicurean_fetal_seising",
    "feature_stylistic_honduran_comprador",
    "feature_hellenistic_scraggly_comfort",
    "feature_intrusive_effluent_hokkaido",
    "feature_unnetted_bay_premillennialist",
    "feature_tossing_denominative_threshing",
    "feature_vestmental_hoofed_transpose",
    "feature_unsurveyed_chopped_feldspathoid",
    "feature_undescribed_methylic_friday",
    "feature_chartered_conceptual_spitting",
    "feature_fustiest_voiced_janet",
    "feature_offshore_defamatory_catalog",
    "feature_jiggish_tritheist_probity",
    "feature_quinsied_increased_braincase",
    "feature_petitionary_evanescent_diallage",
    "feature_fleshly_bedimmed_enfacement",
    "feature_zarathustrian_albigensian_itch",
    "feature_uncompromising_fancy_kyle",
    "feature_seamier_jansenism_inflator",
    "feature_levigate_kindly_dyspareunia",
    "feature_ctenoid_moaning_fontainebleau",
    "feature_unliving_bit_bengaline",
    "feature_culinary_pro_offering",
    "feature_nebule_barmier_bibliomania",
    "feature_unlivable_morbific_traveling",
    "feature_peaty_vulgar_branchia",
    "feature_peculiar_sheenier_quintal",
    "feature_apostate_impercipient_knighthood",
    "feature_collective_stigmatic_handfasting",
    "feature_scenographical_dissentient_trek",
    "feature_unaimed_yonder_filmland",
    "feature_recidivism_petitory_methyltestosterone",
    "feature_unperfect_implemental_cellarage",
    "feature_incommensurable_diffused_curability",
    "feature_bloodied_twinkling_andante",
    "feature_more_hindoo_diageotropism",
    "feature_uncomplimentary_malignant_scoff",
    "feature_sorted_ignitable_sagitta",
    "feature_planar_unessential_bride",
    "feature_reminiscent_unpained_ukulele",
    "feature_whistleable_unbedimmed_chokey",
    "feature_reclaimed_fallibilist_turpentine",
    "feature_tittering_virgilian_decliner",
    "feature_travelled_semipermeable_perruquier",
    "feature_clerkish_flowing_chapati",
    "feature_oversea_permed_insulter",
    "feature_congenital_conched_perithecium",
    "feature_untouchable_unsolvable_agouti",
    "feature_grazed_blameful_desiderative",
    "feature_coordinated_undecipherable_gag",
    "feature_tranquilizing_abashed_glyceria",
    "feature_irresponsive_compositive_ramson",
    "feature_attuned_southward_heckle",
    "feature_supergene_legible_antarthritic",
    "feature_descendent_decanal_hon",
    "feature_encompassing_skeptical_salience",
    "feature_mazy_superrefined_punishment",
    "feature_whitened_remanent_blast",
    "feature_intercalative_helvetian_infirmarian",
    "feature_axillary_reluctant_shorty",
    "feature_liege_unexercised_ennoblement",
    "feature_caespitose_unverifiable_intent",
    "feature_hypothetic_distressing_endemic",
    "feature_calculating_unenchanted_microscopium",
    "feature_zymotic_varnished_mulga",
    "feature_headhunting_unsatisfied_phenomena",
    "feature_voltairean_consolidative_parallel",
    "feature_obeisant_vicarial_passibility",
    "feature_myographic_gawkier_timbale",
    "feature_diverticular_punjabi_matronship",
    "feature_impractical_endorsed_tide",
    "feature_pansophic_merino_pintado",
    "feature_quadratic_untouched_liberty",
    "feature_leaky_maroon_pyrometry",
    "feature_degenerate_officinal_feasibility",
    "feature_continuate_unprocurable_haversine",
    "feature_naval_edified_decarbonize",
    "feature_tonal_graptolitic_corsac",
    "feature_unbreakable_nosological_comedian",
    "feature_iconomatic_boozier_age",
    "feature_unsparred_scarabaeid_anthologist",
    "feature_unextinct_smectic_isa",
    "feature_hotfoot_behaviorist_terylene",
    "feature_crablike_panniered_gloating",
    "feature_untrimmed_monaxial_accompanist",
    "feature_unrequired_waxing_skeptic",
    "feature_buxom_curtained_sienna",
    "feature_uncertified_myrmecological_nagger",
    "feature_chopfallen_fasciate_orchidologist",
    "feature_publishable_apiarian_rollick",
    "feature_fumarolic_known_sharkskin",
    "feature_hawkish_domiciliary_duramen",
    "feature_agronomic_cryptal_advisor",
    "feature_lost_quirky_botel",
    "feature_unnourishing_indiscreet_occiput",
    "feature_galvanometric_sturdied_billingsgate",
    "feature_undisguised_whatever_gaul",
    "feature_generative_honorific_tughrik",
    "feature_venatic_intermetallic_darling",
    "feature_precooled_inoperable_credence",
    "feature_autodidactic_gnarlier_pericardium",
    "feature_glare_factional_assessment",
    "feature_peltate_okay_info",
    "feature_syrian_coital_counterproof",
    "feature_helpable_chanciest_fractionisation",
    "feature_ovular_powered_neckar",
    "feature_reconciling_dauby_database",
    "feature_roiling_trimeric_kurosawa",
    "feature_gossamer_placable_wycliffite",
    "feature_exacerbating_presentationism_apagoge",
    "feature_glyptic_unrubbed_holloway",
    "feature_centric_shaggier_cranko",
    "feature_atlantic_uveal_incommunicability",
    "feature_ruthenian_uncluttered_vocalizing",
    "feature_camphorated_spry_freemartin",
    "feature_built_reincarnate_sherbet",
    "feature_chelonian_pyknic_delphi",
    "feature_demure_groutiest_housedog",
    "feature_loyal_fishy_pith",
    "feature_accessorial_aroused_crochet",
    "feature_communicatory_unrecommended_velure",
    "feature_antipathetical_terrorful_ife",
    "feature_consecrate_legislative_cavitation",
    "feature_retinoscopy_flinty_wool",
    "feature_undetermined_idle_aftergrowth",
    "feature_uretic_seral_decoding",
    "feature_trim_axial_suffocation",
    "feature_questionable_diplex_caesarist",
    "feature_hypogastric_effectual_sunlight",
    "feature_severe_tricky_pinochle",
    "feature_methylated_necrophilic_serendipity",
    "feature_unmoved_alt_spoonerism",
    "feature_unsparing_moralistic_commissary",
    "feature_loricate_cryptocrystalline_ethnology",
    "feature_intended_involute_highbinder",
    "feature_intertwined_leeriest_suffragette",
    "feature_merovingian_tenebrism_hartshorn",
    "feature_rimmed_conditional_archipelago",
    "feature_irritant_euphuistic_weka",
    "feature_armoured_finable_skywriter",
    "feature_subglobular_unsalable_patzer",
    "feature_caecilian_unexperienced_ova",
    "feature_substandard_permissible_paresthesia",
    "feature_scenic_cormophytic_bilirubin",
    "feature_univalve_abdicant_distrail",
    "feature_vizierial_courtlier_hampton",
    "feature_acerb_venusian_piety",
    "feature_periscopic_thirteenth_cartage",
    "feature_resuscitative_communicable_brede",
    "feature_dispiriting_araeostyle_jersey",
    "feature_roasting_slaked_reposition",
    "feature_frequentative_participial_waft",
    "feature_euterpean_frazzled_williamsburg",
    "feature_kerygmatic_splashed_ziegfeld",
    "feature_nubblier_plosive_deepening",
    "feature_springlike_crackjaw_bheesty",
    "feature_hemispherical_unabsolved_aeolipile",
    "feature_synoptic_botryose_earthwork",
    "feature_mucky_loanable_gastrostomy",
    "feature_sudsy_polymeric_posteriority",
    "feature_strychnic_structuralist_chital",
    "feature_patristical_analysable_langouste",
    "feature_nucleophilic_uremic_endogen",
    "feature_highland_eocene_berean",
    "feature_reserved_cleanable_soldan",
    "feature_cerebrovascular_weeny_advocate",
    "feature_invalid_chromatographic_cornishman",
    "feature_groggy_undescried_geosphere",
    "feature_unbeaten_orological_dentin",
    "feature_dipped_sent_giuseppe",
    "feature_salian_suggested_ephemeron",
    "feature_maximal_unobserving_desalinisation",
    "feature_biannual_maleficent_thack",
    "feature_eruptive_seasoned_pharmacognosy",
    "feature_malacological_differential_defeated",
    "feature_massed_nonracial_ecclesiologist",
    "feature_calefactive_anapaestic_jerome",
    "feature_tribal_germinable_yarraman",
    "feature_stereotypic_ebracteate_louise",
    "feature_hexametric_ventricose_limnology",
    "feature_interdental_mongolian_anarchism",
    "feature_agile_unrespited_gaucho",
    "feature_unknown_reusable_cabbage",
    "feature_chuffier_analectic_conchiolin",
    "feature_altern_unnoticed_impregnation",
    "feature_hydrologic_cymric_nyctophobia",
    "feature_beady_unkind_barret",
    "feature_discrepant_ventral_shicker",
    "feature_sallowish_cognisant_romaunt",
    "feature_castrated_presented_quizzer",
    "feature_bushwhacking_unaligned_imperturbability",
    "feature_covalent_unreformed_frogbit",
    "feature_revitalizing_dashing_photomultiplier",
    "feature_flakiest_fleecy_novelese",
    "feature_refreshed_untombed_skinhead",
    "feature_scorbutic_intellectualism_mongoloid",
    "feature_commensurable_industrial_jungfrau",
    "feature_expressed_abhominable_pruning",
    "feature_extractable_serrulate_swing",
    "feature_instrumentalist_extrovert_cassini",
    "feature_alkaline_pistachio_sunstone",
    "feature_plexiform_won_elk",
    "feature_caressive_cognate_cubature",
    "feature_voltairean_dyslogistic_epagoge",
    "feature_leggiest_slaggiest_inez",
    "feature_whopping_eminent_attempter",
    "feature_sixteen_inbreed_are",
    "feature_designer_notchy_epiploon",
    "feature_croupiest_shaded_thermotropism",
    "feature_amygdaloidal_intersectional_canonry",
    "feature_busty_unfitted_keratotomy",
    "feature_log_unregenerate_babel",
    "feature_farcical_spinal_samantha",
    "feature_unburied_exponent_pace",
    "feature_dentilingual_removed_osmometer",
    "feature_arillate_nickelic_hemorrhage",
    "feature_restricted_aggregately_workmanship",
    "feature_ganoid_osiered_mineralogy",
    "feature_introvert_symphysial_assegai",
    "feature_sludgy_implemental_sicily",
    "feature_criticisable_authentical_deprecation",
    "feature_inhabited_pettier_veinlet",
    "feature_confiscatory_triennial_pelting",
    "feature_leisurable_dehortatory_pretoria",
    "feature_favoring_prescript_unorthodoxy",
    "feature_curling_aurorean_iseult",
    "feature_hibernating_soritic_croupe",
    "feature_spookiest_expedite_overnighter",
    "feature_foamy_undrilled_glaciology",
    "feature_uncurtailed_translucid_coccid",
    "feature_limitable_astable_physiology",
    "feature_assenting_darn_arthropod",
    "feature_terrific_epigamic_affectivity",
    "feature_inflexed_lamaism_crit",
    "feature_indentured_communicant_tulipomania",
    "feature_uncharged_unovercome_smolder",
    "feature_decent_solo_stickup",
    "feature_polaroid_squalliest_applause",
    "feature_interrogatory_isohyetal_atacamite",
    "feature_covalent_methodological_brash",
    "feature_undrossy_serpentiform_sack",
    "feature_faltering_tergal_tip",
    "feature_undirected_perdu_ylem",
    "feature_vulcanological_sepulchral_spean",
]

FEATURE_NAMES_SMALL = [
    "feature_unwonted_trusted_fixative",
    "feature_introvert_symphysial_assegai",
    "feature_jerkwater_eustatic_electrocardiograph",
    "feature_canalicular_peeling_lilienthal",
    "feature_unvaried_social_bangkok",
    "feature_crowning_frustrate_kampala",
    "feature_store_apteral_isocheim",
    "feature_haziest_lifelike_horseback",
    "feature_grandmotherly_circumnavigable_homonymity",
    "feature_assenting_darn_arthropod",
    "feature_beery_somatologic_elimination",
    "feature_cambial_bigoted_bacterioid",
    "feature_unaired_operose_lactoprotein",
    "feature_moralistic_heartier_typhoid",
    "feature_twisty_adequate_minutia",
    "feature_unsealed_suffixal_babar",
    "feature_planned_superimposed_bend",
    "feature_winsome_irreproachable_milkfish",
    "feature_flintier_enslaved_borsch",
    "feature_agile_unrespited_gaucho",
    "feature_glare_factional_assessment",
    "feature_slack_calefacient_tableau",
    "feature_undivorced_unsatisfying_praetorium",
    "feature_silver_handworked_scauper",
    "feature_communicatory_unrecommended_velure",
    "feature_stylistic_honduran_comprador",
    "feature_travelled_semipermeable_perruquier",
    "feature_bhutan_imagism_dolerite",
    "feature_lofty_acceptable_challenge",
    "feature_antichristian_slangiest_idyllist",
    "feature_apomictical_motorized_vaporisation",
    "feature_buxom_curtained_sienna",
    "feature_gullable_sanguine_incongruity",
    "feature_unforbidden_highbrow_kafir",
    "feature_chuffier_analectic_conchiolin",
    "feature_branched_dilatory_sunbelt",
    "feature_univalve_abdicant_distrail",
    "feature_exorbitant_myeloid_crinkle",
]

FEATURE_NAMES_MEDIUM = [
    "feature_dichasial_hammier_spawner",
    "feature_rheumy_epistemic_prancer",
    "feature_pert_performative_hormuz",
    "feature_hillier_unpitied_theobromine",
    "feature_perigean_bewitching_thruster",
    "feature_renegade_undomestic_milord",
    "feature_koranic_rude_corf",
    "feature_demisable_expiring_millepede",
    "feature_unscheduled_malignant_shingling",
    "feature_clawed_unwept_adaptability",
    "feature_rubblier_chlorotic_stogy",
    "feature_untumbled_histologic_inion",
    "feature_piffling_inflamed_jupiter",
    "feature_abstersive_emotional_misinterpreter",
    "feature_unluckiest_mulley_benzyl",
    "feature_escutcheoned_timocratic_kotwal",
    "feature_integrated_extroversive_ambivalence",
    "feature_vedic_mitral_swiz",
    "feature_reclaimed_fallibilist_turpentine",
    "feature_gone_honduran_worshipper",
    "feature_insociable_exultant_tatum",
    "feature_outdated_tapered_speciation",
    "feature_leggiest_slaggiest_inez",
    "feature_chaldean_vixenly_propylite",
    "feature_hysteric_mechanized_recklinghausen",
    "feature_glare_factional_assessment",
    "feature_highland_eocene_berean",
    "feature_seemlier_reorient_monandry",
    "feature_expressed_abhominable_pruning",
    "feature_castrated_presented_quizzer",
    "feature_restricted_aggregately_workmanship",
    "feature_scorbutic_intellectualism_mongoloid",
    "feature_telephonic_shakable_bollock",
    "feature_subglobular_unsalable_patzer",
    "feature_syrian_coital_counterproof",
    "feature_supergene_legible_antarthritic",
    "feature_hypothetic_distressing_endemic",
    "feature_torturesome_estimable_preferrer",
    "feature_greedier_favorable_enthymeme",
    "feature_decent_solo_stickup",
    "feature_unmalleable_resistant_kingston",
    "feature_seamier_jansenism_inflator",
    "feature_lordly_lamellicorn_buxtehude",
    "feature_mattery_past_moro",
    "feature_helpable_chanciest_fractionisation",
    "feature_spookiest_expedite_overnighter",
    "feature_culinary_pro_offering",
    "feature_ganoid_osiered_mineralogy",
    "feature_hotfoot_behaviorist_terylene",
    "feature_severe_tricky_pinochle",
    "feature_maximal_unobserving_desalinisation",
    "feature_voltairean_consolidative_parallel",
    "feature_unmoved_alt_spoonerism",
    "feature_tittering_virgilian_decliner",
    "feature_methylated_necrophilic_serendipity",
    "feature_calceolate_pudgy_armure",
    "feature_unsparing_moralistic_commissary",
    "feature_criticisable_authentical_deprecation",
    "feature_undetermined_idle_aftergrowth",
    "feature_antipathetical_terrorful_ife",
    "feature_caressive_cognate_cubature",
    "feature_crablike_panniered_gloating",
    "feature_exacerbating_presentationism_apagoge",
    "feature_epicurean_fetal_seising",
    "feature_casuistic_barbarian_monochromy",
    "feature_haematoid_runaway_nightjar",
    "feature_croupiest_shaded_thermotropism",
    "feature_zarathustrian_albigensian_itch",
    "feature_unsparred_scarabaeid_anthologist",
    "feature_unco_terefah_thirster",
    "feature_sallowish_cognisant_romaunt",
    "feature_grazed_blameful_desiderative",
    "feature_whitened_remanent_blast",
    "feature_introvert_symphysial_assegai",
    "feature_casemated_ibsenian_grantee",
    "feature_intertwined_leeriest_suffragette",
    "feature_built_reincarnate_sherbet",
    "feature_axillary_reluctant_shorty",
    "feature_descendent_decanal_hon",
    "feature_untrimmed_monaxial_accompanist",
    "feature_desiderative_commiserative_epizoa",
    "feature_subdued_spiffier_kano",
    "feature_affricative_bromic_raftsman",
    "feature_doggish_whacking_headscarf",
    "feature_congenital_conched_perithecium",
    "feature_terrific_epigamic_affectivity",
    "feature_enlightening_mirthful_laurencin",
    "feature_continuate_unprocurable_haversine",
    "feature_fustiest_voiced_janet",
    "feature_ovular_powered_neckar",
    "feature_questionable_diplex_caesarist",
    "feature_nubblier_plosive_deepening",
    "feature_revitalizing_dashing_photomultiplier",
    "feature_chafed_undenominational_backstitch",
    "feature_assenting_darn_arthropod",
    "feature_consecrate_legislative_cavitation",
    "feature_inhabited_pettier_veinlet",
    "feature_voltairean_dyslogistic_epagoge",
    "feature_log_unregenerate_babel",
    "feature_agile_unrespited_gaucho",
    "feature_whopping_eminent_attempter",
    "feature_travelled_semipermeable_perruquier",
    "feature_horizontal_snug_description",
    "feature_inhibited_snowiest_drawing",
    "feature_salian_suggested_ephemeron",
    "feature_diverticular_punjabi_matronship",
    "feature_reminiscent_unpained_ukulele",
    "feature_unsurveyed_chopped_feldspathoid",
    "feature_degenerate_officinal_feasibility",
    "feature_undrossy_serpentiform_sack",
    "feature_festering_controvertible_hostler",
    "feature_leukemic_paler_millikan",
    "feature_subapostolic_dungy_fermion",
    "feature_wale_planned_tolstoy",
    "feature_bifacial_hexastyle_hemialgia",
    "feature_revealable_aeonian_elvira",
    "feature_perceivable_gasiform_psammite",
    "feature_smugger_hydroponic_farnesol",
    "feature_apostate_impercipient_knighthood",
    "feature_unperfect_implemental_cellarage",
    "feature_stereotypic_ebracteate_louise",
    "feature_offshore_defamatory_catalog",
    "feature_tribal_germinable_yarraman",
    "feature_dispiriting_araeostyle_jersey",
    "feature_clerkish_flowing_chapati",
    "feature_venatic_intermetallic_darling",
    "feature_nebule_barmier_bibliomania",
    "feature_unknown_reusable_cabbage",
    "feature_planar_unessential_bride",
    "feature_irritant_euphuistic_weka",
    "feature_whistleable_unbedimmed_chokey",
    "feature_roiling_trimeric_kurosawa",
    "feature_loyal_fishy_pith",
    "feature_unlivable_morbific_traveling",
    "feature_untellable_penal_allegorization",
    "feature_coordinated_undecipherable_gag",
    "feature_congealed_lee_steek",
    "feature_intrusive_effluent_hokkaido",
    "feature_epidermic_scruffiest_prosperity",
    "feature_quadratic_untouched_liberty",
    "feature_hypogastric_effectual_sunlight",
    "feature_contused_festal_geochemistry",
    "feature_hexametric_ventricose_limnology",
    "feature_paramagnetic_complex_gish",
    "feature_chuffier_analectic_conchiolin",
    "feature_indefatigable_enterprising_calf",
    "feature_untouchable_unsolvable_agouti",
    "feature_reduplicate_conoid_albite",
    "feature_sixteen_inbreed_are",
    "feature_dovetailed_winy_hanaper",
    "feature_bushwhacking_unaligned_imperturbability",
    "feature_undescribed_methylic_friday",
    "feature_inflexed_lamaism_crit",
    "feature_calefactive_anapaestic_jerome",
    "feature_agronomic_cryptal_advisor",
    "feature_unbreakable_nosological_comedian",
    "feature_unmodernized_vasodilator_galenist",
    "feature_paraffinoid_irreplevisable_ombu",
    "feature_invalid_extortionary_titillation",
    "feature_fake_trident_agitator",
    "feature_wieldable_defiled_aperitive",
    "feature_contaminative_intrusive_tagrag",
    "feature_himyarite_tetragonal_deceit",
    "feature_tossing_denominative_threshing",
    "feature_dendritic_prothallium_sweeper",
    "feature_ruffianly_uncommercial_anatole",
    "feature_hellenistic_scraggly_comfort",
    "feature_squishiest_unsectarian_support",
    "feature_myographic_gawkier_timbale",
    "feature_faltering_tergal_tip",
    "feature_antisubmarine_foregoing_cryosurgery",
    "feature_altern_unnoticed_impregnation",
    "feature_foamy_undrilled_glaciology",
    "feature_appraisive_anagrammatical_tentacle",
    "feature_emmetropic_heraclitean_conducting",
    "feature_collective_stigmatic_handfasting",
    "feature_iridic_unpropertied_spline",
    "feature_farcical_spinal_samantha",
    "feature_urochordal_swallowed_curn",
    "feature_wombed_reverberatory_colourer",
    "feature_merovingian_tenebrism_hartshorn",
    "feature_stylistic_honduran_comprador",
    "feature_caecilian_unexperienced_ova",
    "feature_fumarolic_known_sharkskin",
    "feature_belgravian_salopian_sheugh",
    "feature_zymotic_varnished_mulga",
    "feature_learned_claustral_quiddity",
    "feature_brickier_heterostyled_scrutiny",
    "feature_unnetted_bay_premillennialist",
    "feature_uncurtailed_translucid_coccid",
    "feature_jiggish_tritheist_probity",
    "feature_groggy_undescried_geosphere",
    "feature_hawkish_domiciliary_duramen",
    "feature_base_ingrain_calligrapher",
    "feature_accessorial_aroused_crochet",
    "feature_fierier_goofier_follicle",
    "feature_unburied_exponent_pace",
    "feature_chelonian_pyknic_delphi",
    "feature_tarry_meet_chapel",
    "feature_generative_honorific_tughrik",
    "feature_smoggy_niftiest_lunch",
    "feature_coalier_typhoid_muntin",
    "feature_chopfallen_fasciate_orchidologist",
    "feature_euterpean_frazzled_williamsburg",
    "feature_quinsied_increased_braincase",
    "feature_unlawful_superintendent_brunet",
    "feature_naval_edified_decarbonize",
    "feature_scenic_cormophytic_bilirubin",
    "feature_unvaried_social_bangkok",
    "feature_fissirostral_multifoliate_chillon",
    "feature_gutta_exploitive_simpson",
    "feature_assertive_worsened_scarper",
    "feature_intersubjective_juristic_sagebrush",
    "feature_rowable_unshod_noise",
    "feature_volitional_ascensive_selfhood",
    "feature_barest_kempt_crowd",
    "feature_curtained_gushier_tranquilizer",
    "feature_intermontane_vertical_moo",
    "feature_isotopic_hymenial_starwort",
    "feature_inseminated_filarial_mesoderm",
    "feature_passerine_ultraist_neon",
    "feature_westering_immunosuppressive_crapaud",
    "feature_reported_slimy_rhapsody",
    "feature_juvenalian_paunchy_uniformitarianism",
    "feature_lipogrammatic_blowsier_seismometry",
    "feature_draconic_contractible_romper",
    "feature_indirect_concrete_canaille",
    "feature_puberulent_nondescript_laparoscope",
    "feature_unsurveyed_boyish_aleph",
    "feature_dismaying_chaldean_tallith",
    "feature_epitaxial_loathsome_essen",
    "feature_malagasy_abounding_circumciser",
    "feature_tortured_arsenical_arable",
    "feature_sludgy_implemental_sicily",
    "feature_uretic_seral_decoding",
    "feature_roasting_slaked_reposition",
    "feature_peculiar_sheenier_quintal",
    "feature_publishable_apiarian_rollick",
    "feature_interdental_mongolian_anarchism",
    "feature_reserved_cleanable_soldan",
    "feature_eruptive_seasoned_pharmacognosy",
    "feature_softish_unseparated_caudex",
    "feature_univalve_abdicant_distrail",
    "feature_levigate_kindly_dyspareunia",
    "feature_intercalative_helvetian_infirmarian",
    "feature_centric_shaggier_cranko",
    "feature_unliving_bit_bengaline",
    "feature_misanthropic_knurliest_freebooty",
    "feature_confiscatory_triennial_pelting",
    "feature_peltate_okay_info",
    "feature_reconciling_dauby_database",
    "feature_unspotted_practiced_gland",
    "feature_iconomatic_boozier_age",
    "feature_congenial_transmigrant_isobel",
    "feature_impractical_endorsed_tide",
    "feature_huskiest_compartmental_jacquerie",
    "feature_cairned_fumiest_ordaining",
    "feature_calculating_unenchanted_microscopium",
    "feature_mucky_loanable_gastrostomy",
    "feature_loricate_cryptocrystalline_ethnology",
    "feature_ctenoid_moaning_fontainebleau",
    "feature_armoured_finable_skywriter",
    "feature_intended_involute_highbinder",
    "feature_commensurable_industrial_jungfrau",
    "feature_spagyric_echt_alum",
    "feature_illiterate_stomachal_terpene",
    "feature_demure_groutiest_housedog",
    "feature_dentilingual_removed_osmometer",
    "feature_plexiform_won_elk",
    "feature_induplicate_hoarse_disbursement",
    "feature_vizierial_courtlier_hampton",
    "feature_lost_quirky_botel",
    "feature_hydrologic_cymric_nyctophobia",
    "feature_atlantic_uveal_incommunicability",
    "feature_incommensurable_diffused_curability",
    "feature_midget_noncognizable_plenary",
    "feature_unapplicable_jerkiest_klemperer",
    "feature_unamazed_tumular_photomicrograph",
    "feature_nucleophilic_uremic_endogen",
    "feature_unnourishing_indiscreet_occiput",
    "feature_hypersonic_volcanological_footwear",
    "feature_uncomplimentary_malignant_scoff",
    "feature_interrogatory_isohyetal_atacamite",
    "feature_invalid_chromatographic_cornishman",
    "feature_encompassing_skeptical_salience",
    "feature_beady_unkind_barret",
    "feature_gossamer_placable_wycliffite",
    "feature_flavourful_seismic_erica",
    "feature_petitionary_evanescent_diallage",
    "feature_leisurable_dehortatory_pretoria",
    "feature_chaotic_granitoid_theist",
    "feature_cerebrovascular_weeny_advocate",
    "feature_cyrenaic_unschooled_silurian",
    "feature_uninclosed_handcrafted_springing",
    "feature_indentured_communicant_tulipomania",
    "feature_synoptic_botryose_earthwork",
    "feature_acerb_venusian_piety",
    "feature_vestmental_hoofed_transpose",
    "feature_unrated_intact_balmoral",
    "feature_tonal_graptolitic_corsac",
    "feature_refreshed_untombed_skinhead",
    "feature_communicatory_unrecommended_velure",
    "feature_sudsy_polymeric_posteriority",
    "feature_inexpugnable_gleg_candelilla",
    "feature_headhunting_unsatisfied_phenomena",
    "feature_scenographical_dissentient_trek",
    "feature_trim_axial_suffocation",
    "feature_buxom_curtained_sienna",
    "feature_recidivism_petitory_methyltestosterone",
    "feature_irresponsive_compositive_ramson",
    "feature_more_hindoo_diageotropism",
    "feature_apomictical_motorized_vaporisation",
    "feature_seclusive_emendatory_plangency",
    "feature_discrepant_ventral_shicker",
    "feature_obeisant_vicarial_passibility",
    "feature_unrelieved_rawish_cement",
    "feature_strychnic_structuralist_chital",
    "feature_unaimed_yonder_filmland",
    "feature_designer_notchy_epiploon",
    "feature_favoring_prescript_unorthodoxy",
    "feature_cheering_protonemal_herd",
    "feature_conjugal_postvocalic_rowe",
    "feature_palmy_superfluid_argyrodite",
    "feature_outsized_admonishing_errantry",
    "feature_brawny_confocal_frail",
    "feature_arillate_nickelic_hemorrhage",
    "feature_bloodied_twinkling_andante",
    "feature_rusted_unassisting_menaquinone",
    "feature_biannual_maleficent_thack",
    "feature_extractable_serrulate_swing",
    "feature_covalent_methodological_brash",
    "feature_periscopic_thirteenth_cartage",
    "feature_tranquilizing_abashed_glyceria",
    "feature_faustian_unventilated_lackluster",
    "feature_frequentative_participial_waft",
    "feature_flakiest_fleecy_novelese",
    "feature_liege_unexercised_ennoblement",
    "feature_amygdaloidal_intersectional_canonry",
    "feature_rural_inquisitional_trotline",
    "feature_caespitose_unverifiable_intent",
    "feature_precooled_inoperable_credence",
    "feature_ruthenian_uncluttered_vocalizing",
    "feature_camphorated_spry_freemartin",
    "feature_mined_game_curse",
    "feature_amoebaean_wolfish_heeler",
    "feature_peaty_vulgar_branchia",
    "feature_ratlike_matrilinear_collapsability",
    "feature_attuned_southward_heckle",
    "feature_substandard_permissible_paresthesia",
    "feature_curling_aurorean_iseult",
    "feature_fleshly_bedimmed_enfacement",
    "feature_hendecagonal_deathly_stiver",
    "feature_massed_nonracial_ecclesiologist",
    "feature_instrumentalist_extrovert_cassini",
    "feature_patristical_analysable_langouste",
    "feature_gullable_sanguine_incongruity",
    "feature_phellogenetic_vibrational_jocelyn",
    "feature_autodidactic_gnarlier_pericardium",
    "feature_unbeaten_orological_dentin",
    "feature_resuscitative_communicable_brede",
    "feature_limitable_astable_physiology",
    "feature_chartered_conceptual_spitting",
    "feature_oversea_permed_insulter",
    "feature_busty_unfitted_keratotomy",
    "feature_alkaline_pistachio_sunstone",
    "feature_incitant_trochoidal_oculist",
    "feature_stelar_balmiest_pellitory",
    "feature_hypermetropic_unsighted_forsyth",
    "feature_calycled_living_birmingham",
    "feature_transmontane_clerkly_value",
    "feature_together_suppositive_aster",
    "feature_planned_superimposed_bend",
    "feature_undirected_perdu_ylem",
    "feature_antipodal_unable_thievery",
    "feature_scrobiculate_unexcitable_alder",
    "feature_apophthegmatical_catechetical_millet",
    "feature_fragrant_fifteen_brian",
    "feature_churrigueresque_talc_archaicism",
    "feature_uncompromising_fancy_kyle",
    "feature_retinoscopy_flinty_wool",
    "feature_pansophic_merino_pintado",
    "feature_galvanometric_sturdied_billingsgate",
    "feature_mazy_superrefined_punishment",
    "feature_aztecan_encomiastic_pitcherful",
    "feature_sorted_ignitable_sagitta",
    "feature_padded_peripteral_pericranium",
    "feature_trabeate_eutherian_valedictory",
    "feature_undisguised_whatever_gaul",
    "feature_migrant_reliable_chirurgery",
    "feature_permanent_cottony_ballpen",
    "feature_conceding_ingrate_tablespoonful",
    "feature_vulcanological_sepulchral_spean",
    "feature_unrequired_waxing_skeptic",
    "feature_hibernating_soritic_croupe",
    "feature_unstacked_trackable_blizzard",
    "feature_multilinear_sharpened_mouse",
    "feature_autarkic_constabulary_dukedom",
    "feature_christadelphian_euclidean_boon",
    "feature_covalent_unreformed_frogbit",
    "feature_basaltic_arid_scallion",
    "feature_uncharged_unovercome_smolder",
    "feature_leaky_maroon_pyrometry",
    "feature_polaroid_squalliest_applause",
    "feature_jerkwater_eustatic_electrocardiograph",
    "feature_malacological_differential_defeated",
    "feature_ambisexual_boiled_blunderer",
    "feature_endangered_unthreaded_firebrick",
    "feature_unextinct_smectic_isa",
    "feature_kerygmatic_splashed_ziegfeld",
    "feature_palatalized_unsucceeded_induration",
    "feature_springlike_crackjaw_bheesty",
    "feature_unweary_congolese_captain",
    "feature_uncertified_myrmecological_nagger",
    "feature_hemispherical_unabsolved_aeolipile",
    "feature_glyptic_unrubbed_holloway",
    "feature_rimmed_conditional_archipelago",
    "feature_bleeding_arabesque_pneuma",
    "feature_dipped_sent_giuseppe",
    "feature_undivorced_unsatisfying_praetorium",
    "feature_reclinate_cruciform_lilo",
]

COLUMN_NAMES_LEGACY = FEATURE_NAMES_LEGACY + TARGET_NAMES
COLUMN_NAMES_SMALL = FEATURE_NAMES_SMALL + TARGET_NAMES
COLUMN_NAMES_MEDIUM = FEATURE_NAMES_MEDIUM + TARGET_NAMES



from ._numerai_main import (
    fetch_numerai_training,
    fetch_numerai_test,
    fetch_numerai_validation,
    fetch_numerai_live,
    fetch_numerai_tournament,
    fetch_numerai_example_predictions,
    fetch_numerai_example_validation_predictions,
    fetch_numerai_feature_metadata,
)
from ._numerai_main_meta import (
    TARGET_NAMES_UNIQUE,
    TARGET_NAMES,
    FEATURE_NAMES_LEGACY,
    FEATURE_NAMES_SMALL,
    FEATURE_NAMES_MEDIUM,
    COLUMN_NAMES_LEGACY,
    COLUMN_NAMES_SMALL,
    COLUMN_NAMES_MEDIUM,
)

__all__ = [
    "fetch_numerai_training",
    "fetch_numerai_test",
    "fetch_numerai_validation",
    "fetch_numerai_live",
    "fetch_numerai_tournament",
    "fetch_numerai_example_predictions",
    "fetch_numerai_example_validation_predictions",
    "fetch_numerai_feature_metadata",
    "TARGET_NAMES_UNIQUE",
    "TARGET_NAMES",
    "FEATURE_NAMES_LEGACY",
    "FEATURE_NAMES_SMALL",
    "FEATURE_NAMES_MEDIUM",
    "COLUMN_NAMES_LEGACY",
    "COLUMN_NAMES_SMALL",
    "COLUMN_NAMES_MEDIUM",
]
